@@ -34,7 +34,8 @@ from slixmpp.plugins.base import BasePlugin
 from slixmpp.plugins.xep_0004 import Form  # type: ignore[attr-defined]
 from slixmpp.plugins.xep_0060 import XEP_0060  # type: ignore[attr-defined]
 from slixmpp.plugins.xep_0163 import XEP_0163
-from slixmpp.stanza import Message, Iq
+from slixmpp.roster import RosterNode  # type: ignore[attr-defined]
+from slixmpp.stanza import Iq, Message, Presence
 
 from .base_session_manager import BaseSessionManager, TrustLevel
 
@@ -79,7 +80,7 @@ async def _publish_item_and_configure_node(
     node: str,
     item: ET.Element,
     item_id: str,
-    options: Dict[str, Any]
+    options: Dict[str, str]
 ) -> None:
     """
     Publishes an item and makes sure that the node is configured correctly.
@@ -153,8 +154,22 @@ def _make_session_manager(xmpp: BaseXMPP, xep_0384: "XEP_0384") -> Type[SessionM
                             "pubsub#max_items": "max"
                         }
                     )
-                except Exception as e:
-                    raise BundleUploadFailed(f"Bundle upload failed: {bundle}") from e
+                except Exception:  # pylint: disable=broad-exception-caught
+                    # Try again without MAX_ITEMS set, which is not strictly necessary.
+                    try:
+                        await _publish_item_and_configure_node(
+                            xep_0060,
+                            our_bare_jid,
+                            node,
+                            item,
+                            item_id=str(bundle.device_id),
+                            options={
+                                "pubsub#access_model": "open",
+                                "pubsub#persist_items": "true"
+                            }
+                        )
+                    except Exception as e:
+                        raise BundleUploadFailed(f"Bundle upload failed: {bundle}") from e
 
                 return
 
@@ -172,7 +187,7 @@ def _make_session_manager(xmpp: BaseXMPP, xep_0384: "XEP_0384") -> Type[SessionM
                         options={
                             "pubsub#access_model": "open",
                             "pubsub#persist_items": "true",
-                            "pubsub#max_items": 1
+                            "pubsub#max_items": "1"
                         }
                     )
                 except Exception:  # pylint: disable=broad-exception-caught
@@ -316,7 +331,7 @@ def _make_session_manager(xmpp: BaseXMPP, xep_0384: "XEP_0384") -> Type[SessionM
                     options={
                         "pubsub#access_model": "open",
                         "pubsub#persist_items": "true",
-                        "pubsub#max_items": 1
+                        "pubsub#max_items": "1"
                     }
                 )
             except Exception:  # pylint: disable=broad-exception-caught
@@ -384,6 +399,12 @@ def _make_session_manager(xmpp: BaseXMPP, xep_0384: "XEP_0384") -> Type[SessionM
                     return twomemo.etree.parse_device_list(device_list_elt)
                 if namespace == oldmemo.oldmemo.NAMESPACE:
                     return oldmemo.etree.parse_device_list(device_list_elt)
+            except XMLSchemaValidationError as e:
+                log.warning(
+                    f"Malformed device list for {bare_jid} under namespace {namespace}, treating as empty",
+                    exc_info=e
+                )
+                return {}
             except Exception as e:
                 raise DeviceListDownloadFailed(
                     f"Device list download failed for {bare_jid} under namespace {namespace}"
@@ -478,12 +499,11 @@ async def _prepare(
 
     session_manager = await _make_session_manager(xmpp, xep_0384).create(
         [
-            # Uncomment when SCE is available
-            # twomemo.Twomemo(
-            #     storage,
-            #     max_num_per_session_skipped_keys,
-            #     max_num_per_message_skipped_keys
-            # ),
+            twomemo.Twomemo(
+                storage,
+                max_num_per_session_skipped_keys,
+                max_num_per_message_skipped_keys
+            ),
             oldmemo.Oldmemo(
                 storage,
                 max_num_per_session_skipped_keys,
@@ -523,6 +543,11 @@ class XEP_0384(BasePlugin, metaclass=ABCMeta):  # pylint: disable=invalid-name
     The plugin does not treat the protocol versions as separate encryption mechanisms, instead it manages all
     versions transparently with no manual intervention required.
 
+    Certain initialization tasks such as a data consistency check are transparently ran in the background when
+    the plugin is loaded. The ``omemo_initialized`` event is fired when those initial background tasks are
+    done. Waiting for this event can be useful e.g. in automated testing environments to be sure that a test
+    client has generated and uploaded its OMEMO data before continuing.
+
     Tip:
         A lot of essential functionality is accessible via the `SessionManager` instance that is returned by
         :meth:`get_session_manager`. The session manager is the core of the underlying OMEMO library and
@@ -557,6 +582,8 @@ class XEP_0384(BasePlugin, metaclass=ABCMeta):  # pylint: disable=invalid-name
         xmpp.add_event_handler("twomemo_device_list_publish", self._on_device_list_update)
         xmpp.add_event_handler("oldmemo_device_list_publish", self._on_device_list_update)
 
+        xmpp.add_event_handler("changed_subscription", self._on_subscription_changed)
+
         xep_0163.add_interest(TWOMEMO_DEVICE_LIST_NODE)
         xep_0163.add_interest(OLDMEMO_DEVICE_LIST_NODE)
 
@@ -666,6 +693,7 @@ class XEP_0384(BasePlugin, metaclass=ABCMeta):  # pylint: disable=invalid-name
             session_manager = await self.__session_manager_task
             self.__session_manager = session_manager
             self.__session_manager_task = None
+            self.xmpp.event("omemo_initialized")
             return session_manager
 
         # If the session manager is currently being built, wait for it to be done
@@ -720,6 +748,168 @@ class XEP_0384(BasePlugin, metaclass=ABCMeta):  # pylint: disable=invalid-name
 
         await session_manager.update_device_list(namespace, msg["from"].bare, device_list)
 
+    async def _on_subscription_changed(self, presence: Presence) -> None:
+        """
+        Callback to handle presence subscription changes.
+
+        Args:
+            presence: The presence stanza triggering this callback.
+        """
+
+        jid = JID(presence["from"].bare)
+
+        roster: RosterNode = self.xmpp.client_roster
+
+        pep_enabled = jid in roster and roster[jid]["subscription"] == "both"
+
+        log.debug(f"Subscription changed for {jid}; PEP enabled: {pep_enabled}")
+
+        for namespace in [ twomemo.twomemo.NAMESPACE, oldmemo.oldmemo.NAMESPACE ]:
+            subscribed = (await self.storage.load_primitive(
+                f"/slixmpp/subscribed/{jid}/{namespace}",
+                bool
+            )).maybe(None)
+
+            if subscribed is None:
+                # This JID is not tracked.
+                return
+
+            # Remove manual subscriptions if PEP is enabled now
+            if pep_enabled and subscribed:
+                await self._unsubscribe(namespace, jid)
+
+            # Add a manual subscription if PEP is disabled now
+            if not pep_enabled and not subscribed:
+                await self._subscribe(namespace, jid)
+
+    async def _subscribe(self, namespace: str, jid: JID) -> None:
+        """
+        Manually subscribe to the device list pubsub node of the JID and track the subscription status.
+
+        Args:
+            namespace: The OMEMO version namespace (not the node).
+            jid: The JID whose device list to manually subscribe to. Can be a bare (aka "userhost") JID but
+                doesn't have to.
+        """
+
+        jid = JID(jid.bare)
+
+        log.debug(f"Manually subscribing to {namespace} device list for {jid}")
+
+        node = {
+            twomemo.twomemo.NAMESPACE: TWOMEMO_DEVICE_LIST_NODE,
+            oldmemo.oldmemo.NAMESPACE: OLDMEMO_DEVICE_LIST_NODE
+        }.get(namespace, None)
+
+        if node is None:
+            raise UnknownNamespace(f"Unknown namespace during device list subscription: {namespace}")
+
+        xep_0060: XEP_0060 = self.xmpp["xep_0060"]
+
+        try:
+            await xep_0060.subscribe(jid, node)
+        except IqError as e:
+            # Failure to subscribe is non-critical here, simply debug log the error (and don't update the
+            # subscription status).
+            log.debug(f"Couldn't subscribe to {namespace} device list of {jid.bare}", exc_info=e)
+        else:
+            await self.storage.store(f"/slixmpp/subscribed/{jid.bare}/{namespace}", True)
+
+    async def _unsubscribe(self, namespace: str, jid: JID) -> None:
+        """
+        Manually unsubscribe from the device list pubsub node of the JID and track the subscription status.
+
+        Args:
+            namespace: The OMEMO version namespace (not the node).
+            jid: The JID whose device list to manually unsubscribe from. Can be a bare (aka "userhost") JID
+                but doesn't have to.
+        """
+
+        jid = JID(jid.bare)
+
+        log.debug(f"Manually unsubscribing from {namespace} device list for {jid}")
+
+        node = {
+            twomemo.twomemo.NAMESPACE: TWOMEMO_DEVICE_LIST_NODE,
+            oldmemo.oldmemo.NAMESPACE: OLDMEMO_DEVICE_LIST_NODE
+        }.get(namespace, None)
+
+        if node is None:
+            raise UnknownNamespace(f"Unknown namespace during device list unsubscription: {namespace}")
+
+        xep_0060: XEP_0060 = self.xmpp["xep_0060"]
+
+        try:
+            await xep_0060.unsubscribe(jid, node)
+        except IqError as e:
+            # Don't really care about any of the possible Iq error cases:
+            # https://xmpp.org/extensions/xep-0060.html#subscriber-unsubscribe-error
+            # Worst case we keep receiving updates we don't need.
+            log.debug(f"Couldn't unsubscribe from {namespace} device list of {jid.bare}", exc_info=e)
+
+        await self.storage.store(f"/slixmpp/subscribed/{jid.bare}/{namespace}", False)
+
+    async def refresh_device_lists(self, jids: Set[JID], force_download: bool = False) -> None:
+        """
+        Ensure that up-to-date device lists for the JIDs are cached. This is done automatically by
+        :meth:`encrypt_message`. You don't have to ever manually call this method, but you can do so for
+        optimization reasons. For example, in a UI-based IM application, this method can be called when an
+        OMEMO-enabled chat tab/window is opened, to be optimally prepared if the user decides to send an
+        encrypted message.
+
+        Args:
+            jids: The JIDs whose device lists to refresh. Can be bare (aka "userhost") JIDs but don't have
+                to.
+            force_download: Force downloading the device list even if pubsub/PEP are enabled to automatically
+                keep the cached device lists up-to-date.
+
+        Raises:
+            Exception: all exceptions raised by :meth:`SessionManager.refresh_device_lists` are forwarded
+                as-is.
+        """
+
+        session_manager = await self.get_session_manager()
+        storage = self.storage
+        roster: RosterNode = self.xmpp.client_roster
+
+        for jid in jids:
+            jid = JID(jid.bare)
+
+            if jid.bare == self.xmpp.boundjid.bare:
+                # Skip ourselves
+                continue
+
+            # Track which namespaces require a manual refresh
+            refresh_namespaces: Set[str] = \
+                { twomemo.twomemo.NAMESPACE, oldmemo.oldmemo.NAMESPACE } if force_download else set()
+
+            # PEP is "enabled" with mutual presence subscription and applies to all backends when enabled.
+            pep_enabled = jid in roster and roster[jid]["subscription"] == "both"
+
+            if not pep_enabled:
+                # If PEP is not enabled, check whether manual subscription is enabled instead. Manual
+                # subscription is tracked per-backend.
+                for namespace in [ twomemo.twomemo.NAMESPACE, oldmemo.oldmemo.NAMESPACE ]:
+                    subscribed = (await storage.load_primitive(
+                        f"/slixmpp/subscribed/{jid.bare}/{namespace}",
+                        bool
+                    )).maybe(None)
+
+                    if not subscribed:
+                        # If not subscribed already (or the subscription status is unknown), manually
+                        # subscribe to stay up-to-date automatically in the future. This trusts that servers,
+                        # even if they support multi-subscribe, would not generate exact duplicate
+                        # subscriptions with differing subscription ids.
+                        await self._subscribe(namespace, jid)
+                        refresh_namespaces.add(namespace)
+
+            for namespace in refresh_namespaces:
+                # Force-download the device lists that need a manual refresh
+                try:
+                    await session_manager.refresh_device_list(namespace, jid.bare)
+                except omemo.DeviceListDownloadFailed as e:
+                    log.debug(f"Couldn't manually fetch {namespace} device list, probably doesn't exist: {e}")
+
     async def encrypt_message(
         self,
         stanza: Message,
@@ -732,8 +922,8 @@ class XEP_0384(BasePlugin, metaclass=ABCMeta):  # pylint: disable=invalid-name
 
         Args:
             stanza: The stanza to encrypt.
-            recipient_jids: The JID of the recipients. Can be a bare (aka "userhost") JIDs but doesn't have
-                to. A single JID can be used.
+            recipient_jids: The JID of the recipients. Can be bare (aka "userhost") JIDs but doesn't have to.
+                A single JID can be used.
             identifier: A value that is passed on to :meth:`_devices_blindly_trusted` and
                 :meth:`_prompt_manual_trust` in case a trust decision is required for any of the recipient
                 devices. This value is not processed or altered, it is simply passed through. Refer to the
@@ -772,6 +962,9 @@ class XEP_0384(BasePlugin, metaclass=ABCMeta):  # pylint: disable=invalid-name
         if not recipient_jids:
             raise ValueError("At least one JID must be specified")
 
+        # Make sure all recipient device lists are available
+        await self.refresh_device_lists(recipient_jids)
+
         recipient_bare_jids = frozenset({ recipient_jid.bare for recipient_jid in recipient_jids })
 
         # Prepare the plaintext for all protocol versions
@@ -891,7 +1084,7 @@ class XEP_0384(BasePlugin, metaclass=ABCMeta):  # pylint: disable=invalid-name
 
         if message.namespace == twomemo.twomemo.NAMESPACE:
             # Do SCE unpacking here
-            raise NotImplementedError("SCE not supported yet")
+            raise NotImplementedError(f"SCE not supported yet. Plaintext: {plaintext}")
 
         if message.namespace == oldmemo.oldmemo.NAMESPACE:
             stanza = copy(stanza)
